/*! ********************************************
 * MidiController.java
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.util;
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;
	
public class MidiController implements IGrayBoxModule 
{
// MidiController ---
	protected int midiOutPorts = 0xFFFF;
	protected MidiEvent[] _eventBuffer = null;
	protected int _eventBufferSize = 64;
	protected int _usedEvents = 0;
	
	public void clearMIDIEventBuffer()
	{
		for (int i = 0; i < _eventBufferSize; i++)_eventBuffer[i].clear();
		_usedEvents = 0;
	}
	public int getEventBufferSize(){ return _eventBufferSize; }
	public int setEventBufferSize(int size)
	{
		if (size == _eventBufferSize) return _eventBufferSize;
		if (size < 64)size = 64;
		_eventBuffer = new MidiEvent[size];
		for (int i = 0; i < size; i++)_eventBuffer[i] = new MidiEvent();
		_usedEvents = 0;
		return _eventBufferSize = size;
	}
// MIDI Controlle
	public void noteOff(int ch, int key){ noteOff(ch, key, 0); }
	public void noteOff(int ch, int key, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0x80;
		_ev.channel = ch;
		_ev.value0 = key;
	}
	public void noteOn(int ch, int key, int velo){ noteOn(ch, key, velo, 0); }
	public void noteOn(int ch, int key, int velo, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0x90;
		_ev.channel = ch;
		_ev.value0 = key;
		_ev.value1 = velo;
	}
	public void keyAfterTouch(int ch, int key, int val){ keyAfterTouch(ch, key, val, 0); }
	public void keyAfterTouch(int ch, int key, int val, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xA0;
		_ev.channel = ch;
		_ev.value0 = key;
		_ev.value1 = val;
	}
	public void controlChange(int ch, int ccn, int val){ controlChange(ch, ccn, val, 0); }
	public void controlChange(int ch, int ccn, int val, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xB0;
		_ev.channel = ch;
		_ev.value0 = ccn;
		_ev.value1 = val;
	}
	public void programChange(int ch, int val){ programChange(ch, val, 0); }
	public void programChange(int ch, int val, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xC0;
		_ev.channel = ch;
		_ev.value0 = val;
	}
	public void channelAfterTouch(int ch, int val){ channelAfterTouch(ch, val, 0); }
	public void channelAfterTouch(int ch, int val, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xD0;
		_ev.channel = ch;
		_ev.value0 = val;
	}
	public void pitchBendChange(int ch, int lsb, int msb){ pitchBendChange(ch, lsb, msb, 0); }
	public void pitchBendChange(int ch, int lsb, int msb, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xE0;
		_ev.channel = ch;
		_ev.value0 = lsb;
		_ev.value1 = msb;
	}
	public void pitchBendChangeW(int ch, int val){ pitchBendChange(ch, val, 0); }
	public void pitchBendChangeW(int ch, int val, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xE0;
		_ev.channel = ch;
		_ev.value0 = val & 0x7f;
		_ev.value1 = val >> 7 & 0x7f;
	}
	public void systemExclusive(byte[] msg, int dataSize){ systemExclusive(msg, dataSize, 0); }
	public void systemExclusive(byte[] msg, int dataSize, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xF0;
		_ev.channel = 0;
		_ev.data = msg;
		_ev.dataSize = dataSize;
	}
	public void midiMessage(int type, int ch, int val0, int val1){ midiMessage(type, ch, val0, val1, null, 0, 0); }
	public void midiMessage(int type, int ch, int val0, int val1, byte[] data, int dataSize, int offsetFrames)
	{
		if (_usedEvents >= _eventBufferSize) return;
		MidiEvent _ev = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = type;
		_ev.channel = ch;
		_ev.value0 = val0;
		_ev.value1 = val1;
		_ev.data = data;
		_ev.dataSize = dataSize;
	}
// Constructor
	public MidiController(){ this(64); }
	public MidiController(int eventBufferSize)
	{
		if (eventBufferSize < 16)eventBufferSize = 16;
		_eventBufferSize = eventBufferSize;
		_eventBuffer = new MidiEvent[_eventBufferSize];
		for (int i = 0; i < _eventBufferSize; i++) _eventBuffer[i] = new MidiEvent();
		_usedEvents = 0;
	}
// Interface - IfGrayBoxModule -----
	public void updateSampleRate(float sampleRate){}
	public void updateAudioChannels(int ch){}
	public void updateProcessBlockSize(int size){}
	
	public void resetModule(){ clearMIDIEventBuffer(); }
	public void initiModule(){ clearMIDIEventBuffer(); }
	public void connectModule(GrayBox host){ clearMIDIEventBuffer(); }
	public void disconnectModule(){ clearMIDIEventBuffer(); }
	public void updateBypassStatus(boolean bypass){ clearMIDIEventBuffer(); }
	
	public void startProcess(){ clearMIDIEventBuffer(); }
	public void stopProcess(){ clearMIDIEventBuffer(); }
	public void suspendProcess(){ clearMIDIEventBuffer(); }
	public void resumeProcess(){ clearMIDIEventBuffer(); }
	public void process(ProcessData procData)
	{
		for (int i = 0; i < _usedEvents; i++)
		{
			MidiEvent _ev = _eventBuffer[i];
			MidiEvent _list = procData.midiEventList;
			while (_list.next != null)
			{
				if (_ev.offsetFrames >= _list.next.offsetFrames) _list = _list.next;
				else break;
			}
			_ev.next = _list.next;
			_list.next = _ev;
		}
		_usedEvents = 0;
	}
}